/* This helps keep the overlay scrollbar away from the border a bit
 * as far as damage areas are concerned. It doesn't actually move the
 * slider at all. Doing so increases the chance that fractional scaling
 * will not result in the damage overlapping a border pixel cascading
 * into an alpha blend by the compositor.
 *
 * Remove this if/when libadwaita does the same.
 *
 * See GNOME/libadwaita#800
 */
scrollbar.overlay-indicator.vertical range.vertical trough slider {
  outline-width: 0;
}

vte-terminal.padded {
  margin: 0 0 1px 0;
  padding: 7px 6px 3px 6px;
}

vte-terminal:drop(active) { border-radius: 0 0 11px 11px; }

window.maximized vte-terminal:drop(active),
window.tiled-bottom vte-terminal:drop(active) { border-radius: 0; }

vte-terminal revealer.size label
{
  padding: 3px 12px;
  border-radius: 8px;
  margin: 4px;
}

findbar .close-button
{
  min-height: 0;
  min-width: 0;
  border-radius: 99px;
}

findbar .close-button image
{
  padding: 0;
}

popover button.pill.flat
{
  padding: 0;
}

themeselector {
  margin: 9px;
}
themeselector checkbutton {
  padding: 0;
  min-height: 44px;
  min-width: 44px;
  padding: 1px;
  background-clip: content-box;
  border-radius: 9999px;
  box-shadow: inset 0 0 0 1px @borders;
}
themeselector checkbutton.follow:checked,
themeselector checkbutton.light:checked,
themeselector checkbutton.dark:checked {
  box-shadow: inset 0 0 0 2px var(--accent-bg-color);
}
themeselector checkbutton.follow {
  background-image: linear-gradient(to bottom right, #fff 49.99%, #202020 50.01%);
}
themeselector checkbutton.light {
  background-color: #fff;
}
themeselector checkbutton.dark {
  background-color: #202020;
}
themeselector checkbutton radio {
  -gtk-icon-source: none;
  border: none;
  background: none;
  box-shadow: none;
  min-width: 12px;
  min-height: 12px;
  transform: translate(27px, 14px);
  padding: 2px;
}
window checkbutton.theme-selector radio:checked {
  -gtk-icon-source: -gtk-icontheme("object-select-symbolic");
  background-color: var(--accent-bg-color);
  color: var(--accent-fg-color);
}

button.palette {
  border-radius: 12px;
  box-shadow: 0 0 0 1px @borders;
  margin: 2px;
  padding: 0;
}
button.palette.selected {
  box-shadow: 0 0 0 2px var(--accent-bg-color);
}
button.palette image {
  border-radius: 99px;
  background-color: transparent;
  color: transparent;
  padding: 2px;
  margin: -3px -3px 0 0;
}
button.palette.selected image {
  background-color: var(--accent-bg-color);
  color: var(--accent-fg-color);
}
palettepreview > box {
  padding: 12px 12px;
}
palettepreviewcolor {
  border-radius: 5px;
  min-width: 20px;
  min-height: 16px;
}
preferencesgroup flowboxchild {
  outline-offset: 2px;
  border-radius: 12px;
  padding: 0;
}
preferencesgroup flowboxchild button.palette:not(.selected) {
  box-shadow: 0 0 0 1px alpha(black, 0.03),
              0 1px 3px 1px alpha(black, .07),
              0 2px 6px 2px alpha(black, .03);
}

.card columnview {
  background: transparent;
}

popover.tab-menu contents {
  padding: 0;
}
popover.tab-menu listview {
  background: transparent;
  padding: 6px;
}
popover.tab-menu listview row {
  border-radius: 9px;
  min-height: 32px;
  padding: 0;
  margin: 0;
}
popover.tab-menu listview row:hover {
  background: alpha(currentColor, .13);
}
popover.tab-menu listview row:active {
  background: alpha(currentColor, .19);
}
popover.tab-menu listview header {
  padding: 0;
  margin: 0;
  min-height: 0;
}
popover.tab-menu listview header > * {
  padding: 12px 6px 6px 12px;
}
popover.tab-menu listview row > * {
  padding: 0 12px;
}

/* Keep padding within visual bell area, not outside it */
fullscreenbox > toolbarview > revealer.top-bar > windowhandle > box {
  padding-top: 0;
  padding-bottom: 0;
}
.visual-bell-area {
  padding-top: 0px;
  padding-bottom: 0px;
}
window.has-tab-bar .visual-bell-area {
  padding-top: 3px;
  padding-bottom: 3px;
}
